// ============================================================================
//    Author: Kenneth Perkins
//    Date:   Sept 4, 2020
//    Taken From: http://programmingnotes.org/
//    File:  Modal.js
//    Description: Javascript that handles modal popup related functions
//    Example:
//        // Show Modal
//        Modal.show('selector')
//
//        // Hide Modal
//        Modal.hide('selector')
// ============================================================================
/**
* NAMESPACE: Modal
* USE: Handles Modal related functions
*/
var Modal = Modal || {};
(function(namespace) {
    'use strict';

// -- Public data --
    // Property to hold public variables and functions
    let exposed = namespace;

    // Set class names and other shared data
    const settings = {
        handleTypeResize: 'resize',
        handleTypeDrag: 'drag',

        // Element class names
        classNameModal: '.modal',
        classNameModalPanel: '.modal-panel',
        classNameCloseButton: '.modal-close-button',
        classNameModalHeader: '.modal-header',
        classNameModalHeaderText: '.modal-header-text',
        classNameModalBody: '.modal-body',
        classNameModalFooter: '.modal-footer',
        classNameModalFooterText: '.modal-footer-text',
        classNameDraggable: '.modal-draggable',
        classNameDraggableIcon: '.icon',
        classNameShow: '.show',
        classNameResizable: '.modal-resizable',
        classNameResizableIcon: '.icon',
        classNameDialog: '.dialog',
        classNameButton: '.modal-button',
        classNameButtonCancel: '.cancel',
        classNameButtonOK: '.ok',
        classNameButtonContainer: '.buttonContainer',
        classNameResizeMargin: '.resizeMargin',

        // Element data names
        dataNameCloseOnOuterClick: 'data-closeOnOuterClick',
        dataNameDraggable: 'data-draggable',
        dataNameIsDraggable: 'data-isDraggable',
        dataNameResizable: 'data-resizable',
        dataNameIsResizable: 'data-isResizable',
        cleanClassName: (str) => {
            return str ? str.trim().replace('.', '') : '';
        },
    };

    exposed.settings = settings;

    /**
    * FUNCTION: init
    * USE: Initializes the modal button clicks.
    * @param element: JavaScript element to search for modals.
    * @return: N/A.
    */
    exposed.init = (element = document) => {
        addClickEvents(element);
    }

    /**
    * FUNCTION: show
    * USE: Shows the modal popup.
    * @param element: JavaScript element/String selector of the modal to show.
    * @return: N/A.
    */
    exposed.show = (element) => {
        element = verifyElement(element);
        addClass(element, settings.classNameShow);
    }

    /**
    * FUNCTION: hide
    * USE: Hides the modal popup.
    * @param element: JavaScript element/String selector of the modal to hide.
    * @return: N/A.
    */
    exposed.hide = (element) => {
        element = verifyElement(element);
        resetPosition(element);
        removeClass(element, settings.classNameShow);
    }

    /**
    * FUNCTION: getPanel
    * USE: Returns the modal panel element.
    * @param element: JavaScript element/String selector of the modal.
    * @return: The modal panel element.
    */
    exposed.getPanel = (element) => {
        element = verifyElement(element);
        return element.querySelector(settings.classNameModalPanel);
    }

    /**
    * FUNCTION: getHeader
    * USE: Returns the modal header element.
    * @param element: JavaScript element/String selector of the modal.
    * @return: The modal header element.
    */
    exposed.getHeader = (element) => {
        element = verifyElement(element);
        return element.querySelector(settings.classNameModalHeader);
    }

    /**
    * FUNCTION: getHeaderText
    * USE: Returns the modal header text element.
    * @param element: JavaScript element/String selector of the modal.
    * @return: The modal header text element.
    */
    exposed.getHeaderText = (element) => {
        element = verifyElement(element);
        return element.querySelector(settings.classNameModalHeaderText);
    }

    /**
    * FUNCTION: getBody
    * USE: Returns the modal body element.
    * @param element: JavaScript element/String selector of the modal.
    * @return: The modal body element.
    */
    exposed.getBody = (element) => {
        element = verifyElement(element);
        return element.querySelector(settings.classNameModalBody);
    }

    /**
    * FUNCTION: getFooter
    * USE: Returns the modal footer element.
    * @param element: JavaScript element/String selector of the modal.
    * @return: The modal footer element.
    */
    exposed.getFooter = (element) => {
        element = verifyElement(element);
        return element.querySelector(settings.classNameModalFooter);
    }

    /**
    * FUNCTION: getFooterText
    * USE: Returns the modal footer text element.
    * @param element: JavaScript element/String selector of the modal.
    * @return: The modal footer text element.
    */
    exposed.getFooterText = (element) => {
        element = verifyElement(element);
        return element.querySelector(settings.classNameModalFooterText);
    }

    /**
    * FUNCTION: showDialog
    * USE: Shows a modal dialog box with an OK/Cancel button.
    * @param options: An object of initialization options.
    *   Its made up of the following properties:
    *   {
    *       title: The title of the dialog
    *       body: The content body of the dialog
    *       onShow(): Optional. Function that allows to do something before show.
    *          Return false to stop showing process
    *       onHide(): Optional. Function that allows to do something before hide.
    *          Return false to stop hidding process
    *       onDestroy(): Optional. Function that allows to do something before
    *          destroy. Return false to stop destroying process
    *       closeOnOuterClick: Optional. Determines if the dialog should close
    *          on outer dialog click. Default is False
    *       draggable: Optional. Determines if the dialog is draggable.
    *          Default is False
    *       resizable: Optional. Determines if the dialog is resizable.
    *          Default is False
    *       autoOpen: Optional. Determines if the dialog should auto open.
    *          Default is True
    *       draggable: Optional. Determines if the dialog is draggable.
    *          Default is False
    *       destroyOnHide: Optional. Determines if the dialog should be removed
    *          from DOM after hiding. Default is True
    *       buttons: Optional. Object allows to add custom buttons to the dialog
    *   }
    * @return: The modal dialog object.
    */
    exposed.showDialog = (options) => {
        options = verifyOptions(options);
        let dialog = createDialog(options);
        if (toBoolean(options.autoOpen)) {
            dialog.show();
        }
        return dialog;
    }

// -- Private data --
    let verifyOptions = (options) => {
        if (typeof options !== 'object') {
            let body = options;
            options = {};
            options.body = body;
        }
        return options;
    }

    let createDialog = (options) => {
        let defaultButtons = {
            cancel: {
                text: 'Cancel',
                cssClass: settings.classNameButtonCancel,
                visible: true,
            },
            ok: {
                text: 'OK',
                cssClass: settings.classNameButtonOK,
                visible: true,
            },
        };

        copyDefaults(options, defaultButtons);

        let modalButtons = options.buttons;
        let container = document.body;
        let modal = createElement('section', [settings.classNameModal, settings.classNameDialog]);
        let modalPanel = createElement('article', [settings.classNameModalPanel, settings.classNameDialog]);
        let modalHeader = createElement('header', [settings.classNameModalHeader, settings.classNameDialog]);
        let modalHeaderText = createElement('div', [settings.classNameModalHeaderText, settings.classNameDialog]);
        let modalBody = createElement('div', [settings.classNameModalBody, settings.classNameDialog]);
        let modalFooter = createElement('footer', [settings.classNameModalFooter, settings.classNameDialog]);
        let buttonContainer = createElement('div', [settings.classNameButtonContainer]);

        modal.appendChild(modalPanel);
        modalPanel.appendChild(modalHeader);
        modalHeader.appendChild(modalHeaderText);
        modalPanel.appendChild(modalBody);
        modalPanel.appendChild(modalFooter);
        modalFooter.appendChild(buttonContainer);

        let showHeaderCloseButton = true;
        let keys = Object.keys(modalButtons);
        keys.forEach((prop, index) => {
            prop = prop.toLowerCase();
            let buttonOptions = modalButtons[prop];
            if (!isNull(buttonOptions.visible) && !toBoolean(buttonOptions.visible)) {
                return false;
            }

            let text = buttonOptions.text || '';
            let cssClass = buttonOptions.cssClass || '';
            let button = createElement('button', [settings.classNameButton, settings.classNameDialog, cssClass]);
            button.innerHTML = !isNull(text) && isEmpty(text) ? '&nbsp;' : text;
            buttonContainer.appendChild(button)

            registerEvents(button, ['click'], (eventClick) => {
                if (buttonOptions.onClick) {
                    let value = buttonOptions.onClick.call(button, eventClick);
                    if (!isNull(value) && !value) {
                        return;
                    }
                }
                hide(options.destroyOnHide);
            });
            if ((index + 1) === keys.length) {
                if (toBoolean(options.resizable)) {
                    addClass(button, settings.classNameResizeMargin);
                }
            }
            showHeaderCloseButton = false;
        });

        if (showHeaderCloseButton) {
            let modalHeaderClose = createElement('div', [settings.classNameCloseButton, settings.classNameDialog]);
            modalHeader.appendChild(modalHeaderClose);
            registerEvents(modalHeaderClose, ['click'], (eventClick) => {
                hide(options.destroyOnHide);
            });
        }

        modalHeaderText.innerHTML = options.title || '';
        modalBody.innerHTML = options.body || '';

        modal.id = `modal_${randomFromTo(1271991, 7281987)}`;
        if (toBoolean(options.closeOnOuterClick)) {
            modal.setAttribute(settings.dataNameCloseOnOuterClick, options.closeOnOuterClick);
        }
        if (toBoolean(options.draggable)) {
            modal.setAttribute(settings.dataNameDraggable, options.draggable);
        }
        if (toBoolean(options.resizable)) {
            modal.setAttribute(settings.dataNameResizable, options.resizable);
        }
        container.appendChild(modal);

        options.autoOpen = !isNull(options.autoOpen) ? options.autoOpen : true;
        if (!options.autoOpen) {
            if (isNull(options.destroyOnHide)) {
                options.destroyOnHide = false;
            }
        }
        options.destroyOnHide = !isNull(options.destroyOnHide) ? options.destroyOnHide : true;

        let show = () => {
            if (options.onShow) {
                let value = options.onShow.call(this);
                if (!isNull(value) && !value) {
                    return;
                }
            }
            setTimeout(()=> {
                exposed.show(modal);
            }, 10);
        }
        let hide = (destroyModal = false) => {
            if (options.onHide) {
                let value = options.onHide.call(this);
                if (!isNull(value) && !value) {
                    return;
                }
            }
            exposed.hide(modal);
            if (destroyModal) {
                setTimeout(()=> {
                    destroy();
                }, 500);
            }
        }
        let destroy = () => {
            if (options.onDestroy) {
                let value = options.onDestroy.call(this);
                if (!isNull(value) && !value) {
                    return;
                }
            }
            if (!isDestroyed()) {
                container.removeChild(modal);
            }
        }
        let isDestroyed = () => {
            return !container.contains(modal);
        }
        let isShowing = () => {
            return hasClass(modal, settings.classNameShow);
        }

        addModalClickEvents(modal);

        if (getCoords(modalFooter).position === 'absolute') {
            modalFooter.style.position = 'relative';
            let coords = getCoords(modalFooter);
            modalPanel.style['min-width'] = coords.width + 'px';
            modalBody.style['padding-bottom'] = coords.height + 'px';
            modalFooter.style.position = null;
        }

        return {
            modal: modal,
            show: () => {
                show();
            },
            hide: () => {
                hide(options.destroyOnHide);
            },
            destroy: () => {
                hide(true);
            },
            isDestroyed: () => {
                return isDestroyed();
            },
            isShowing: () => {
                return isShowing();
            }
        };
    }

    let copyDefaults = (options, defaultButtons) => {
        if (isNull(options.buttons)) {
            options.buttons = defaultButtons
            return;
        }
        for (let prop in options.buttons) {
            prop = prop.toLowerCase();
            if (!defaultButtons.hasOwnProperty(prop)) {
                continue;
            }
            options.buttons[prop] = Object.assign(defaultButtons[prop], options.buttons[prop])
        }
    }

    let createElement = (type, classes) => {
        let element = document.createElement(type);
        if (!isNull(classes)) {
            classes.forEach(item => {
               addClass(element, item);
            });
        }
        return element;
    }

    let addClickEvents = (element) => {
        // Add misc click events for the modals
        let modals = element.querySelectorAll(settings.classNameModal);
        modals.forEach((modal, index) => {
            addModalClickEvents(modal, index);
        });
    }

    let addModalClickEvents = (modal, index = 0) => {
        let registerCloseOnOuterClick = false;

        // Add button clicks for the close button
        let closeButton = modal.querySelector(settings.classNameCloseButton);
        if (!isNull(closeButton)) {
            registerEvents(closeButton, ['click'], closeButtonEvent);
        }
        // Add functionalty to make the modal draggable
        if (shouldDrag(modal) && !isDraggable(modal)) {
            // Get elements and make sure they exist
            let panel = exposed.getPanel(modal);
            if (isNull(panel)) {
                throw new TypeError(`Unable to make Modal #${index + 1} draggable. Reason: Modal has no 'Modal Panel' associated with it.`);
            }
            // If a header exists, attach a drag handle to it
            let handle = getHandle(modal, settings.handleTypeDrag);
            if (isNull(handle)) {
                let header = exposed.getHeader(modal);
                if (!isNull(header)) {
                    handle = createHandle(header, settings.handleTypeDrag);
                } else {
                    addClass(panel, settings.classNameDraggable);
                }
            }
            makeDraggable({
                element: panel,
                handle: handle,
                keepInViewPort: true,
            });
            // Mark that this modal is draggable
            setIsDraggable(modal, true)
        }

        // Add functionalty to make the modal resizable
        if (shouldResize(modal) && !isResizable(modal)) {
            // Get elements and make sure they exist
            let panel = exposed.getPanel(modal);
            if (isNull(panel)) {
                throw new TypeError(`Unable to make Modal #${index + 1} resizable. Reason: Modal has no 'Modal Panel' associated with it.`);
            }
            // If a footer exists, attach a resize handle to it
            let handle = getHandle(modal, settings.handleTypeResize);
            if (isNull(handle)) {
                let footer = exposed.getFooter(modal);
                if (!isNull(footer)) {
                    handle = createHandle(footer, settings.handleTypeResize);
                } else {
                    addClass(panel, settings.classNameResizable);
                }
            }
            makeResizable({
                element: panel,
                handle: handle,
                keepInViewPort: true,
            });
            // Mark that this modal is resizable
            setIsResizable(modal, true);
        }
        // Check to see if we need to register the close on click event
        if (!registerCloseOnOuterClick) {
            registerCloseOnOuterClick = shouldCloseOnOuterClick(modal);
        }

        // Determine if the modal should close if the outer area is clicked
        if (registerCloseOnOuterClick) {
            registerEvents(document, ['click'], backgroundHideEvent);
        }
    }

    let registerEvents = (element, eventNames, func) => {
        eventNames.forEach((eventName, index) => {
            element.removeEventListener(eventName, func);
            element.addEventListener(eventName, func);
        });
    }

    let closeButtonEvent = (event) => {
        let closeButton = event.target || event.currentTarget;
        if (isNull(closeButton)) {
            return;
        }
        let modal = closeButton.closest(settings.classNameModal);
        exposed.hide(modal);
    }

    let backgroundHideEvent = (event) => {
        let target = event.target;
        if (isNull(target)
            || !isElement(target)
            || !hasClass(target, settings.classNameModal)
            || !hasClass(target, settings.classNameShow)
            || !shouldCloseOnOuterClick(target)) {
            return;
        }
        exposed.hide(target);
    }

    let createHandle = (container, type) => {
        let handle = document.createElement('div');
        switch (type) {
            case settings.handleTypeResize:
                addClass(handle, settings.classNameResizable);
                addClass(handle, settings.classNameResizableIcon);
                break;
            case settings.handleTypeDrag:
                addClass(handle, settings.classNameDraggable);
                addClass(handle, settings.classNameDraggableIcon);
                break;
            default:
                throw new TypeError(`Unknown handle type: ${type}`);
                break;
        }
        handle.setAttribute('tabindex', 0);
        container.appendChild(handle);
        return handle;
    }

    let getHandle = (element, type)  => {
        let selector = null;
        switch (type) {
            case settings.handleTypeResize:
                selector = `${settings.classNameResizable}${settings.classNameResizableIcon}`;
                break;
            case settings.handleTypeDrag:
                selector = `${settings.classNameDraggable}${settings.classNameDraggableIcon}`;
                break;
            default:
                throw new TypeError(`Unknown handle type: ${type}`);
                break;
        }
        return element.querySelector(selector);
    }

    let makeResizable = (options) => {
        let resizable = options.element || options;
        let handle = options.handle || resizable;
        let keepInViewPort = !isNull(options.keepInViewPort) ? options.keepInViewPort : false;

        let coords = getCoords(resizable);
        let initialWidth = coords.width;
        let initialHeight = coords.height;

        let mouseDown = (event) => {
            event.stopPropagation();
            // Get the mouse down position of the element
            let coords = getCoords(resizable);
            let offsetX = event.clientX - coords.width;
            let offsetY = event.clientY - coords.height;

            let moveElement = (event) => {
                // Get the new position
                let newWidth = (event.clientX - offsetX);
                let newHeight = (event.clientY - offsetY);

                // Make sure the width/height is never less than the initial values
                newWidth = Math.max(newWidth, initialWidth);
                newHeight = Math.max(newHeight, initialHeight);

                setPosition(resizable, {width: newWidth, height: newHeight});

                if (keepInViewPort && !isInViewport(resizable)) {
                    let bounding = resizable.getBoundingClientRect();
                    let viewInfo = getViewportInfo();
                    let viewportHeight = viewInfo.height;
                    let viewportWidth = viewInfo.width;
                    if (bounding.bottom > viewportHeight) {
                        newHeight -= Math.abs(bounding.bottom - viewportHeight);
                    }
                    if (bounding.right > viewportWidth) {
                        newWidth -= Math.abs(bounding.right - viewportWidth);
                    }
                    setPosition(resizable, {width: newWidth, height: newHeight});
                }
            }
            let reset = (event) => {
                document.removeEventListener('mousemove', moveElement);
                document.removeEventListener('mouseup', reset);
            }
            document.addEventListener('mouseup', reset);
            document.addEventListener('mousemove', moveElement);
        }
        addTouchSupport(handle);
        handle.addEventListener('mousedown', mouseDown);
    }

    let makeDraggable = (options) => {
        let draggable = options.element || options;
        let handle = options.handle || draggable;
        let keepInViewPort = !isNull(options.keepInViewPort) ? options.keepInViewPort : false;

        let mouseDown = (event) => {
            // Get the mouse down position of the element
            let coords = getCoords(draggable);
            let offsetX = event.clientX - coords.left;
            let offsetY = event.clientY - coords.top;

            let moveElement = (event) => {
                // Get the new position
                let newLeft = event.clientX - offsetX;
                let newTop = event.clientY - offsetY;

                setPosition(draggable, {left: newLeft, top: newTop});

                if (keepInViewPort && !isInViewport(draggable)) {
                    let bounding = draggable.getBoundingClientRect();
                    let viewInfo = getViewportInfo();
                    let viewportHeight = viewInfo.height;
                    let viewportWidth = viewInfo.width;

                    if (bounding.left < 0) {
                        newLeft += Math.abs(bounding.left);
                    }
                    if (bounding.top < 0) {
                        newTop += Math.abs(bounding.top);
                    }
                    if (bounding.bottom > viewportHeight) {
                        newTop -= Math.abs(bounding.bottom - viewportHeight);
                    }
                    if (bounding.right > viewportWidth) {
                        newLeft -= Math.abs(bounding.right - viewportWidth);
                    }
                    setPosition(draggable, {left: newLeft, top: newTop});
                }
            }
            let reset = (event) => {
                document.removeEventListener('mousemove', moveElement);
                document.removeEventListener('mouseup', reset);
            }
            document.addEventListener('mouseup', reset);
            document.addEventListener('mousemove', moveElement);
        }
        addTouchSupport(handle);
        handle.addEventListener('mousedown', mouseDown);
    }

    // Check to make sure the element will be within our viewport boundary
    let isInViewport = (element) => {
        let bounding = element.getBoundingClientRect();
        let viewInfo = getViewportInfo();
        return (
            bounding.top >= 0
            && bounding.left >= 0
            && bounding.bottom <= viewInfo.height
            && bounding.right <= viewInfo.width
        );
    }

    let getViewportInfo = () => {
        return {
            height: Math.max(window.innerHeight || 0, document.documentElement.clientHeight || 0),
            width: Math.max(window.innerWidth || 0, document.documentElement.clientWidth || 0),
        };
    }

    let getCoords = (element) => {
        let computedStyle = window.getComputedStyle(element);
        let intProps = ['left', 'top', 'width', 'height'];
        let result = {};
        for (const prop of intProps) {
            result[prop] = parseInt(computedStyle[prop], 10) || parseInt(element.style[prop], 10) || 0;
        }
        let stringProps = ['position'];
        for (const prop of stringProps) {
            result[prop] = (computedStyle[prop] || element.style[prop]).toLowerCase();
        }
        if (result.position !== 'relative') {
            if (!result.left) {
                result.left = element.offsetLeft || 0;
            }
            if (!result.top) {
                result.top = element.offsetTop || 0;
            }
            if (!result.width) {
                result.width = element.offsetWidth || 0;
            }
            if (!result.height) {
                result.height = element.offsetHeight || 0;
            }
        }
        return result;
    }

    let addTouchSupport = (element) => {
        let events = ['touchstart', 'touchmove', 'touchend', 'touchcancel'];
        registerEvents(element, events, touchConverter);
    }

    let touchConverter = (event) => {
        // stop touch event
        event.stopPropagation();
        event.preventDefault();
        let touches = event.changedTouches;
        if (!touches || touches.length < 1) {
            return;
        }
        let first = touches[0];
        let target = first.target;
        let type = '';
        switch (event.type) {
            case 'touchstart':
                type = 'mousedown';
                break;
            case 'touchmove':
                type = 'mousemove';
                break;
            case 'touchend': case 'touchcancel':
                type = 'mouseup';
                break;
            default:
                return;
        }
        let simulatedEvent = null;
        try {
            simulatedEvent = new MouseEvent(type, {
                bubbles: true,
                cancelable: true,
                view: window,
                screenX: first.screenX,
                screenY: first.screenY,
                clientX: first.clientX,
                clientY: first.clientY,
            });
        } catch (e) {
            simulatedEvent = document.createEvent('MouseEvent');
            simulatedEvent.initMouseEvent(type, true, true, window, 1,
                                      first.screenX, first.screenY,
                                      first.clientX, first.clientY, false,
                                      false, false, false, 0 /*left*/, null);
        }
        target.dispatchEvent(simulatedEvent);
    }

    let resetPosition = (element) => {
        let panel = exposed.getPanel(element);
        if (isNull(panel)) {
            return;
        }
        setPosition(panel, {left: null, top: null, width: null, height: null});
    }

    let setPosition = (element, options) => {
        for (const prop in options) {
            if (prop in element.style) {
                let value = options[prop];
                element.style[prop] = !isNull(value) ? parseInt(value, 10) + 'px' : value;
            }
        }
    }

    let shouldCloseOnOuterClick = (element) => {
        let value = element.getAttribute(settings.dataNameCloseOnOuterClick);
        return toBoolean(value);
    }

    let shouldDrag = (element) => {
        let value = element.getAttribute(settings.dataNameDraggable);
        return toBoolean(value);
    }

    let setIsDraggable = (element, value) => {
        element.setAttribute(settings.dataNameIsDraggable, value);
    }

    let isDraggable = (element) => {
        let value = element.getAttribute(settings.dataNameIsDraggable);
        return toBoolean(value);
    }

    let shouldResize = (element) => {
        let value = element.getAttribute(settings.dataNameResizable);
        return toBoolean(value);
    }

    let setIsResizable = (element, value) => {
        element.setAttribute(settings.dataNameIsResizable, value);
    }

    let isResizable = (element) => {
        let value = element.getAttribute(settings.dataNameIsResizable);
        return toBoolean(value);
    }

    let verifyElement = (element) => {
        if (isString(element)) {
            element = document.querySelector(element);
        }
        if (isNull(element)) {
            throw new TypeError(`Unable to process. Element provided is null`);
        } else if (!isElement(element)) {
            throw new TypeError(`Unable to process element of type: ${typeof element}. Reason: '${element}' is not an HTMLElement.`);
        }
        return element;
    }

    let addClass = (element, cssClass) => {
        cssClass = settings.cleanClassName(cssClass);

        let modified = false;
        if (cssClass.length > 0 && !hasClass(element, cssClass)) {
            element.classList.add(cssClass)
            modified = true;
        }
        return modified;
    }

    let removeClass = (element, cssClass) => {
        cssClass = settings.cleanClassName(cssClass);

        let modified = false;
        if (cssClass.length > 0 && hasClass(element, cssClass)) {
            element.classList.remove(cssClass);
            modified = true;
        }
        return modified;
    }

    let hasClass = (element, cssClass) => {
        cssClass = settings.cleanClassName(cssClass);
        return element.classList.contains(cssClass);
    }

    let isString = (item) => {
        return 'string' === typeof item;
    }

    let isNull = (item) => {
        return undefined === item || null === item;
    }

    let isElement = (item) => {
        let value = false;
        try {
            value = item instanceof HTMLElement || item instanceof HTMLDocument;
        }
        catch (e) {
            value = (typeof item==="object") &&
                (item.nodeType===1) && (typeof item.style === "object") &&
                (typeof item.ownerDocument ==="object");
        }
        return value;
    }

    let toBoolean = (value) => {
        value = String(value).trim().toLowerCase();
        let ret = false;
        switch (value) {
            case 'true': case 'yes': case '1':
                ret = true;
                break;
        }
        return ret;
    }

    let isEmpty = (str) => {
        return isNull(str) || String(str).trim().length < 1;
    }

    let randomFromTo = (from, to) => {
        return Math.floor(Math.random() * (to - from + 1) + from);
    }
    
    (function (factory) {
        if (typeof define === 'function' && define.amd) {
            define([], factory);
        } else if (typeof exports === 'object') {
            module.exports = factory();
        }
    }(function() { return namespace; }));     
}(Modal)); // http://programmingnotes.org/