// ============================================================================
//    Author: Kenneth Perkins
//    Date:   Sept 9, 2020
//    Taken From: http://programmingnotes.org/
//    File:  Placeholder.js
//    Description: Module that adds a floating placeholder label to an element
//    Example:
//        // Add Placeholder
//        Placeholder.float('selector')
//
//        // Remove Placeholder
//        Placeholder.remove('selector')
// ============================================================================
/**
* NAMESPACE: Placeholder
* USE: Handles Placeholder related functions
*/
var Placeholder = Placeholder || {};
(function(namespace) {
    'use strict';

// -- Public data --
    // Property to hold public variables and functions
    let exposed = namespace;

    // Set class names and other shared data
    const settings = {
        attributePlaceholder: 'placeholder',

        // Element class names
        classNameFloatingItem: '.floating-placeholder',
        classNameContainer: '.floating-placeholder-container',
        classNamePlaceholder: '.floating-placeholder-label',
        classNameActive: '.active',

        // Element data names
        dataNamePlaceholderCache: 'data-placeholder-cache',
        cleanClassName: (str) => {
            return str ? str.trim().replace('.', '') : '';
        },
    };

    exposed.settings = settings;

    /**
    * FUNCTION: float
    * USE: Adds a floating placeholder to the selected elements.
    * @param element: One or more JavaScript element to add floating placeholder to.
    * @return: N/A.
    */
    exposed.float = (element) => {
        let items = verifyElements(element);
        for (const item of items) {
            addPlaceholder(item);
        }
    }

    /**
    * FUNCTION: remove
    * USE: Removes a floating placeholder from the selected elements.
    * @param element: One or more JavaScript element to remove floating placeholder from.
    * @return: N/A.
    */
    exposed.remove = (element) => {
        let items = verifyElements(element);
        for (const item of items) {
            removePlaceholder(item);
        }
    }

// -- Private data --
    let removePlaceholder = (element) => {
        let container = getContainer(element);
        let parentNode = container.parentNode;
        removeClass(element, settings.classNameFloatingItem);

        let placeholderText = getPlaceholderTextCache(element);
        setPlaceholderText(element, placeholderText);
        setPlaceholderTextCache(element, null);

        getEvents().forEach(eventInfo => {
            eventInfo.names.forEach(eventName => {
                element.removeEventListener(eventName, eventInfo.callback);
            });
        });

        parentNode.insertBefore(element, container);
        parentNode.removeChild(container);
    }

    let addPlaceholder = (element) => {
        addClass(element, settings.classNameFloatingItem);
        let placeholderText = verifyPlaceholderText(element);
        let container = getContainer(element) || createContainer(element);
        let placeholder = getPlaceholder(element) || createPlaceholder(container);

        placeholder.innerHTML = placeholderText;
        placeholder.htmlFor = element.id;

        getEvents().forEach(eventInfo => {
            registerEvents(element, eventInfo.names, eventInfo.callback);
        });

        if (!isEmpty(element.value)) {
            floatPlaceholder(element);
        } else {
            resetPlaceholder(element);
        }
        
        setDetectChangeHandler(element);
    }
    
    let setDetectChangeHandler = (element, property = 'value') => {
        let elementPrototype = Object.getPrototypeOf(element);
        if (!elementPrototype.hasOwnProperty(property)) {
            return;
        }
        let descriptor = Object.getOwnPropertyDescriptor(elementPrototype, property);
        let newProps = {
            get: function () {
                return descriptor.get.apply(element, arguments);
            },
            set: function (t) {
                setTimeout( () => { element.dispatchEvent(new Event(property + 'change')); }, 10);
                return descriptor.set.apply(element, arguments);
            }
        };
        Object.defineProperty(element, property, newProps);
    }

    let floatPlaceholder = (element) => {
        let placeholder = getPlaceholder(element);
        if (!isNull(placeholder)) {
            addClass(placeholder, settings.classNameActive);
        }
    }

    let resetPlaceholder = (element) => {
        let placeholder = getPlaceholder(element);
        if (!isNull(placeholder)) {
            if (isEmpty(element.value)) {
                removeClass(placeholder, settings.classNameActive);
            }
        }
    }

    let getContainer = (element) => {
        return element.closest(settings.classNameContainer);
    }

    let getPlaceholder = (element) => {
        let container = getContainer(element);
        return !isNull(container)
            ? container.querySelector(settings.classNamePlaceholder) : null;
    }

    let createPlaceholder = (container) => {
        let placeholder = document.createElement('label');
        addClass(placeholder, settings.classNamePlaceholder);
        container.appendChild(placeholder);
        return placeholder;
    }

    let createContainer = (element) => {
        let container = document.createElement('div');
        let parentNode = element.parentNode;
        addClass(container, settings.classNameContainer);
        parentNode.insertBefore(container, element);
        container.appendChild(element);
        return container;
    }

    let verifyPlaceholderText = (element) => {
        if (isEmpty(element.id)) {
            element.id = `floating_placeholder_${randomFromTo(1271991, 7281987)}`;
        }
        let placeholderText = getPlaceholderText(element);
        if (!isEmpty(placeholderText)) {
            setPlaceholderTextCache(element, placeholderText);
        }
        setPlaceholderText(element, null);
        return getPlaceholderTextCache(element);
    }

    let setPlaceholderText = (element, value) => {
        addData(element, {key: settings.attributePlaceholder, value: value});
    }

    let getPlaceholderText = (element) => {
        let value = getData(element, settings.attributePlaceholder);
        return value;
    }

    let setPlaceholderTextCache = (element, value) => {
        addData(element, {key: settings.dataNamePlaceholderCache, value: value});
    }

    let getPlaceholderTextCache = (element) => {
        let value = getData(element, settings.dataNamePlaceholderCache);
        return value;
    }

    let floatPlaceholderEvent = (event) => {
        let element = event.currentTarget;
        let nodeName = element.nodeName.toLowerCase();
        switch (event.type) {
            case 'focus':
                if (nodeName === 'select') {
                    return;
                }
                break;
            case 'keyup':
                if (nodeName === 'select') {
                    if (isEmpty(element.value)) {
                        return;
                    }
                } else {
                    return;
                }
                break;
        }
        floatPlaceholder(element);
    }

    let resetPlaceholderEvent = (event) => {
        let element = event.currentTarget;
        let nodeName = element.nodeName.toLowerCase();
        switch (event.type) {
            case 'keyup':
                if (nodeName === 'select') {
                    if (!isEmpty(element.value)) {
                        return;
                    }
                } else {
                    return;
                }
                break;
        }
        resetPlaceholder(element);
    }

    let registerEvents = (element, eventNames, func) => {
        eventNames.forEach((eventName, index) => {
            element.removeEventListener(eventName, func);
            element.addEventListener(eventName, func);
        });
    }

    let addClass = (element, cssClass) => {
        cssClass = settings.cleanClassName(cssClass);

        let modified = false;
        if (cssClass.length > 0 && !hasClass(element, cssClass)) {
            element.classList.add(cssClass)
            modified = true;
        }
        return modified;
    }

    let removeClass = (element, cssClass) => {
        cssClass = settings.cleanClassName(cssClass);

        let modified = false;
        if (cssClass.length > 0 && hasClass(element, cssClass)) {
            element.classList.remove(cssClass);
            modified = true;
        }
        return modified;
    }

    let hasClass = (element, cssClass) => {
        cssClass = settings.cleanClassName(cssClass);
        return element.classList.contains(cssClass);
    }

    let isNull = (item) => {
        return undefined === item || null === item;
    }

    let isEmpty = (str) => {
        return isNull(str) || String(str).trim().length < 1;
    }

    let randomFromTo = (from, to) => {
        return Math.floor(Math.random() * (to - from + 1) + from);
    }

    let addData = (element, data) => {
        if (isNull(data)) {
            return;
        } else if (!Array.isArray(data)) {
            data = [data];
        }
        data.forEach(item => {
            if (!isNull(item.value)) {
                element.setAttribute(item.key, item.value);
            } else {
                removeData(element, item);
            }
        });
    }

    let removeData = (element, data) => {
        if (isNull(data)) {
            return;
        } else if (!Array.isArray(data)) {
            data = [data];
        }
        data.forEach(item => {
            let key = item.key || item;
            element.removeAttribute(key);
        });
    }

    let getData = (element, data) => {
        if (isNull(data)) {
            return null;
        } else if (!Array.isArray(data)) {
            data = [data];
        }
        let results = [];
        data.forEach(item => {
            let key = item.key || item;
            results.push(element.getAttribute(key));
        });
        return results.length == 1 ? results[0] : results;
    }
    
    let isString = (item) => {
        return 'string' === typeof item;
    }    

    let verifyElements = (items) => {
        if (isNull(items)) {
            throw new TypeError('No elements specified');
        } else if (isString(items)) {
            items = document.querySelectorAll(items);
        }        
        if (!isArrayLike(items)) {
            items = [items];
        }
        return items;
    }

    // see if it looks and smells like an iterable object, and do accept length === 0
    let isArrayLike = (item) => {
        return (
            Array.isArray(item) ||
            (!!item &&
                typeof item === "object" &&
                typeof (item.length) === "number" &&
                (item.length === 0 ||
                    (item.length > 0 &&
                        (item.length - 1) in item)
                )
            )
        );
    }

    let getEvents = () => {
        return [
            {names: ['click', 'focus', 'keyup', 'valuechange'], callback: floatPlaceholderEvent},
            {names: ['blur', 'focusout', 'keyup', 'valuechange'], callback: resetPlaceholderEvent},
        ];
    }
    
    (function (factory) {
        if (typeof define === 'function' && define.amd) {
            define([], factory);
        } else if (typeof exports === 'object') {
            module.exports = factory();
        }
    }(function() { return namespace; }));     
}(Placeholder)); // http://programmingnotes.org/