// ============================================================================
//    Author: Kenneth Perkins
//    Date:   Jun 27, 2020
//    Taken From: http://programmingnotes.org/
//    File:  Sort.by.js
//    Description: Namespace which handles complex array sorting. This allows 
//       SQL/LINQ like sorting capabilities with multiple sorting conditions
//       for arrays of any type.
//    Example:
//        // Sort a simple array in descending order
// 		let sorted = Sort.by( [28,91,10,87,19,32], 
// 		[
//           { value:(x) => x, desc: true }
//           // Custom compare example: 
//           // { value:(x) => String(x), compare: (x, y) => y.localeCompare(x) }
// 		]);
// ============================================================================ 
/**
* NAMESPACE: Sort
* USE: Handles complex array sorting with multiple sorting conditions.
*/
var Sort = Sort || {};
(function(namespace) {
    'use strict';	
    
// -- Public data --
    // Property to hold public variables and functions
    let exposed = namespace;
   
    /**
    * FUNCTION: by
    * USE: Sorts an array in ascending or descending order depending
    *   on the sorting conditions. This method returns a new, sorted
    *   array. The original array is left untouched
    * @param arry: The array to be sorted.
    * @param conditions: The conditions by which the array should be sorted.
    *    Accepts multiple sorting conditions. The array is sorted in the 
    *    order in which the conditions were placed (FIFO)
    * @return: The sorted array according to the sorting conditions.
    */
    exposed.by = (arry, conditions = null) => {
        // Make sure the array is valid
        verifyArray(arry);
        
        // Make sure the sorting conditions are in the expected format
        conditions = verifyConditions(conditions);

        // Create a seperate array that holds only the computed values we want to sort.
        // This is done to reduce overhead in the .sort function, so item
        // values are only computed once, as opposed to possibly multiple times
        // for each comparison.
        let sortable = mapArray(arry, conditions);

        // Get sorting info from the sorting conditions
        let sortingInfo = getSortingInfo(conditions);

        // Sort the sortable array according to its pre-computed value
        sortable.sort((x, y) => {
            // Evaluate each condition based on the order they were supplied in
            let evaluation = evaluate(sortingInfo, x, y);
            return evaluation;
        });	

        // Using the sorted sortable array, reorder the original 
        // array by its index and save it into a result list
        let result = sortable.map((item) => {
            return arry[item.index];
        });	

        return result;
    }

    /**
    * FUNCTION: defaultCompare
    * USE: Default comparer that defines the sort order.
    * @param x: The first element for comparison.
    * @param y: The second element for comparison.
    * @param desc: Indicates if comparison should be ascending or descending.
    * @return: The result value of the comparison.
    */
    exposed.defaultCompare = (x, y, desc) => {
        // Default: both values left unchanged
        let evaluation = 0;

        // Check to see if both values are the same
        if (x === y) {
            evaluation = 0;
        }
        // Null values sort after anything else
        else if (isNull(x) || isNull(y)) {
            if (isNull(x)) {
                evaluation = 1;
            } else {
                evaluation = -1;
            }
        } else {
            // x goes before y
            if (x > y) { 
                evaluation = 1
            } // x goes after y
            else if (x < y) {
                evaluation = -1;
            }
            // Reverse order if descending
            if (desc) {
                evaluation *= -1;
            }
        }
        return evaluation;
    }    
    
// -- Private data --
    /**
    * FUNCTION: evaluate
    * USE: Evaluates each sorting condition and determines the 
    *   order the item should be in.
    * @param sortingInfo: The sorting info from the sorting conditions.
    * @param x: The first element for comparison.
    * @param y: The second element for comparison.
    * @return: The sorting order of this item.
    */
    let evaluate = (sortingInfo, x, y) => {
        let retValue = 0;
        // Evaluate each condition based on the order they were supplied in
        for (let index = 0; index < sortingInfo.length; ++index) {
            let desc = sortingInfo[index].desc;
            let compare = sortingInfo[index].compare;
            let valueX = x.values[index].value;
            let valueY = y.values[index].value;

            // Call the comparison function to evaluate the condition.
            // Exit if we find a valid ordering
            let evaluation = compare.call(this, valueX, valueY, desc);
            if (evaluation) {
                retValue = evaluation;
                break;
            }
        }		
        return retValue;
    }

    /**
    * FUNCTION: mapArray
    * USE: Creates a seperate array that holds only the computed 
    *   values we want to sort. This is done to limit overhead, so 
    *   the item values are only computed once in the .sort  
    *   function, as opposed to possibly multiple times for each 
    *   comparison
    * @param arry: The original array to be sorted
    * @param conditions: The sorting conditions
    * @return: The sortable array holding only the computed values
    */    
    let mapArray = (arry, conditions) => {
        let mapped = [];
        for (let index = 0; index < arry.length; ++index) {
            const item = arry[index];
            // Create a new object that represents the item to sort
            let sortableItem = createSortableItem(item, index, conditions);
            mapped.push(sortableItem);            
        }
        return mapped;
    }

    /**
    * FUNCTION: createSortableItem
    * USE: Creates the item that holds the computed values to be sorted,
    *   and saves its array order (index) of the original array  
    * @param item: The array item in the original array
    * @param index: The index the item resides at in the original array
    * @param conditions: The sorting conditions
    * @return: The sortable item holding the computed values
    */     
    let createSortableItem = (item, index, conditions) => {
        let sortableItem = {};
        sortableItem.index = index;
        sortableItem.values = [];

        // Go through each condition and process its value
        Array.prototype.forEach.call(conditions, (condition, index) => {
            // Check to see if there is a custom value format specified
            let formatValue = condition.value;
            if (isNull(formatValue)) {
                if (!isNull(condition.prop) && item.hasOwnProperty(condition.prop)) {
                    formatValue = (item) => item[condition.prop];
                } else { 
                    formatValue = (item) => item;
                }
            }
            if (!isFunction(formatValue)) {
                throw new TypeError(`Unable to determine the value on sorting condition #${index + 1}. Reason: '${formatValue}' is not a function`);
            }

            // Evaluate the field and save its value
            sortableItem.values.push({
                value: formatValue.call(this, item)
            });
        });
        return sortableItem; 
    } 

    /**
    * FUNCTION: getSortingInfo
    * USE: Gets sorting information about the sorting condition
    * @param conditions: Sorting conditions
    * @return: The sorting direction and comparison 
    *   function to be used for each sorting condition.
    */ 
    let getSortingInfo = (conditions) => {
        let sortingInfo = [];

        Array.prototype.forEach.call(conditions, (condition, index) => {
            // Check to see if this condition should be in descending order.
            // Default is ascending sort
            let desc = (!isNull(condition.desc)) ? condition.desc : false;

            // Get the comparison function
            let compare = condition.compare || exposed.defaultCompare;
            if (!isFunction(compare)) { 
                throw new TypeError(`Invalid comparison of type '${typeof compare}' found on sorting condition #${index + 1}. Reason: '${compare}' is not a function`);
            }
            
            // Save the sorting condition info
            sortingInfo.push({
                desc: desc,
                compare: compare
            });
        });
        
        return sortingInfo;
    }

    /**
    * FUNCTION: verifyArray
    * USE: Make sure the array is valid for sorting
    * @param arry: The array to be sorted
    * @return: N/A.
    */    
   let verifyArray = (arry) => {
       if (isNull(arry)) {
           throw new TypeError('Unable to sort. Reason: Array is not defined');
        } else if (!isArrayLike(arry) && !isString(arry)) {
            throw new TypeError(`Unable to sort value of type: ${typeof arry}. Reason: '${arry}' is not an array.`);
        }
    }
    
    /**
    * FUNCTION: verifyConditions
    * USE: Make sure the sorting conditions are in the expected format
    * @param conditions: Sorting conditions
    * @return: The verified sorting condition.
    */    
    let verifyConditions = (conditions) => {
        if (isNull(conditions)) {
            conditions = {};
        }
        if (!isArrayLike(conditions)) { 
            conditions = [conditions];
        }
        return conditions;
    }

    let isFunction = (item) => {
        return 'function' === typeof item
    }

    let isNull = (item) => {
        return undefined === item || null === item
    }

    let isString = (item) => {
        return 'string' == typeof item;
    }
    
    // see if it looks and smells like an iterable object, and do accept length === 0
    let isArrayLike = (item) => {
        return (
            Array.isArray(item) || 
            (!!item &&
                typeof item === "object" &&
                typeof (item.length) === "number" && 
                (item.length === 0 ||
                    (item.length > 0 && 
                        (item.length - 1) in item)
                )
            )
        );
    }   

    (function (factory) {
        if (typeof define === 'function' && define.amd) {
            define([], factory);
        } else if (typeof exports === 'object') {
            module.exports = factory();
        }
    }(function() { return namespace; }));     
 }(Sort)); // http://programmingnotes.org/