// ============================================================================
//    Author: Kenneth Perkins
//    Date:   Jun 8, 2020
//    Taken From: http://programmingnotes.org/
//    File:  Collapsible.js
//    Description: Demonstrates how to display a simple collapsible
//        accordion panel to the page.
// ============================================================================
/**
* NAMESPACE: Collapsible
* USE: Collapsible accordion panel.
*/
var Collapsible = Collapsible || {};
(function(namespace) {
    'use strict';

// -- Public data --
    // Property to hold public variables and functions
    let exposed = namespace;

    // Set class names and other shared data
    const settings = {
        classNameExpanded: '.expanded',
        classNameHeader: '.collapsible-header',
        classNameHeaderText: '.collapsible-header-text',
        classNameBody: '.collapsible-body',
        classNameButton: '.collapsible-button',
        classNameHideContent: '.hide',
        classNameCollapsible: '.collapsible',
        dataNameSlideUp: 'data-slideup',
        dataNameSlideDown: 'data-slidedown',
        dataNameExpanded: 'data-expanded',
        dataNameCurrentlyExpanded: 'data-isCurrentlyExpanded',
        cleanClassName: function(str) {
            return str ? str.trim().replace('.', '') : '';
        },
    };

    exposed.settings = settings;

    /**
    * FUNCTION: init
    * USE: Initializes the collapsible accordion panel button clicks.
    * @param element: JavaScript element to search for collapsible panels
    * @return: N/A.
    */
    exposed.init = (element = document) => {
        addClickEvents(element);
    }

    /**
    * FUNCTION: animate
    * USE: Handles opening/closing of a collapsible accordion panel.
    * @param collapsible: JavaScript element that raised the event that
    *   is decorated with the 'collapsible' class tag.
    * @return: N/A.
    */
    exposed.animate = (collapsible) => {
        if (!collapsible) {
            return;
        }

        // Find the collapsible header row
        // and make sure its found
        let header = collapsible.closest(settings.classNameHeader);
        if (!header) {
            throw new Error('Unable to find header row!');
        }

        // Find the associated collapsible body text
        // and make sure its found
        let body = header.nextElementSibling;
        if (!body) {
            throw new Error('Unable to find content body!');
        }

        // Determine if the content should be expanded or not
        let status = getPanelStatus(body);

        // Get the slide up/down speeds
        let slideUpDuration = status.slideUpDuration;
        let slideDownDuration = status.slideDownDuration;

        // Get the current collapsible status
        let isCurrentlyExpanded = status.isCurrentlyExpanded;

        // Find the action button so we can change its icon
        let button = header.querySelector(settings.classNameButton);

        // Update contents depending on if the row is expanded or not
        if (toBoolean(isCurrentlyExpanded)) {
            // Mark the header row as 'active'
            addClass(header, settings.classNameExpanded);

            // Change the button icon to '-'
            addClass(button, settings.classNameExpanded);

            // Slide down the body
            slideDown(body, slideDownDuration);
        } else {
            // Remove the header row as 'active'
            removeClass(header, settings.classNameExpanded);

            // Change the button icon to '+'
            removeClass(button, settings.classNameExpanded);

            // Slide up the body
            slideUp(body, slideUpDuration);
        }

        // Save the expanded status
        body.setAttribute(settings.dataNameCurrentlyExpanded, isCurrentlyExpanded);
    }

    /**
    * FUNCTION: getHeaders
    * USE: Returns the collapsible headers that are descendants of the object on
    *   which this method was called.
    * @param element: JavaScript element to search for collapsible headers
    * @return: Collapsible headers descendants found in the given element.
    */
    exposed.getHeaders = (element = document) => {
        return element.querySelectorAll(settings.classNameHeader);
    }

    /**
    * FUNCTION: getBodies
    * USE: Returns the collapsible bodies that are descendants of the object on
    *   which this method was called.
    * @param element: JavaScript element to search for collapsible bodies
    * @return: Collapsible body descendants found in the given element.
    */
    exposed.getBodies = (element = document) => {
        return element.querySelectorAll(settings.classNameBody);
    }

    /**
    * FUNCTION: getPanels
    * USE: Returns the collapsible panels (header & bodies) that are
    *   descendants of the object on which this method was called.
    * @param element: JavaScript element to search for collapsible panels
    * @return: Collapsible panel descendants found in the given element.
    */
    exposed.getPanels = (element = document) => {
        let panels = [];

        // Get collapsible headers and bodies
        const headers = exposed.getHeaders(element);
        const bodies = exposed.getBodies(element);

        // Get the maximum item count
        const maxItems = Math.max(headers.length, bodies.length);

        // Go through each header and body and create an
        // object to group them together and save them to a list
        for(let x = 0; x < maxItems; ++x) {
            let header = (x < headers.length) ? headers[x] : null;
            let body = (x < bodies.length) ? bodies[x] : null;

            panels.push({
                header: header,
                body: body,
            });
        }

        return panels;
    }

// -- Private data --

    /**
    * FUNCTION: addClickEvents
    * USE: Adds slider animation button click events for the radio buttons.
    * @param element: JavaScript element to search for collapsibles
    * @return: N/A.
    */
    let addClickEvents = (element = document) => {
        // Initialize each row to its initial
        // opened/closed state, depending on user values
        element.querySelectorAll(settings.classNameCollapsible).forEach(function(collapsible, index) {
            exposed.animate(collapsible);

            // Add button click
            collapsible.addEventListener('click', function(eventClick) {
                eventClick.stopPropagation();
                exposed.animate(this)
            });
        });
    }

    let addClass = (element, cssClass) => {
        cssClass = settings.cleanClassName(cssClass);

        let modified = false;
        if (cssClass.length > 0 && !hasClass(element, cssClass)) {
            element.classList.add(cssClass)
            modified = true;
        }
        return modified;
    }

    let removeClass = (element, cssClass) => {
        cssClass = settings.cleanClassName(cssClass);

        let modified = false;
        if (cssClass.length > 0 && hasClass(element, cssClass)) {
            element.classList.remove(cssClass);
            modified = true;
        }
        return modified;
    }

    let hasClass = (element, cssClass) => {
        cssClass = settings.cleanClassName(cssClass);
        return element.classList.contains(cssClass);
    }

    let slideUp = (body, slideUpDuration) => {
        // Add the class that removes the content body
        addClass(body, settings.classNameHideContent);

        // Set the slide up duration
        body.style.transitionDuration = (null != slideUpDuration) ? slideUpDuration + 'ms' : null;

        // Remove the content body custom height so the transition can take effect
        body.style.height = null;
    }

    let slideDown = (body, slideDownDuration) => {
        // Remove the class that hides the content body
        removeClass(body, settings.classNameHideContent);

        // Set the slide down duration
        body.style.transitionDuration = (null != slideDownDuration) ? slideDownDuration + 'ms' : null;

        // Get the content body height
        let bodyHeight = body.scrollHeight;

        // Get the style for the element so we can get its max height
        let computedStyle = window.getComputedStyle(body);

        // Check to see if the current content body is greater than the maximum allowed height.
        // Scrollbars appear when setting a custom height.
        // We do this check so scroll bars arent visible if they dont need to be
        if (computedStyle && computedStyle.maxHeight) {
            let computedMaxHeight = parseInt(computedStyle.maxHeight, 10);
            if (bodyHeight > computedMaxHeight) {
                // Body height is bigger than the max height, remove the custom overflow value
                // and use its computed overflow style
                body.style.overflow = null;
            } else {
                // Override and set overflow to hidden so scroll bars arent visible
                // if they dont need to be
                body.style.overflow = 'hidden';
            }
        }

        // Set the content body custom height so the transition can take effect
        body.style.height = bodyHeight + 'px';
    }

    let getPanelStatus = (body) => {
        // Get the slide up/down speeds
        let slideUpDuration = body.getAttribute(settings.dataNameSlideUp);
        let slideDownDuration = body.getAttribute(settings.dataNameSlideDown);

        // Get the current collapsible status
        let isCurrentlyExpanded = body.getAttribute(settings.dataNameCurrentlyExpanded);

        // If the current status hasnt been defined yet, use the user defined value
        if (null == isCurrentlyExpanded) {
            isCurrentlyExpanded = body.getAttribute(settings.dataNameExpanded)

            // Assume the row is closed if the user did not define a value
            if (null == isCurrentlyExpanded) {
                isCurrentlyExpanded = false;
            }

            // Remove the delay so the row is immediately opened/closed
            slideUpDuration = 0;
            slideDownDuration = 0;
        } else {
            // The status has been defined in the past. Change its state
            if (toBoolean(isCurrentlyExpanded)) {
                isCurrentlyExpanded = false;
            } else {
                isCurrentlyExpanded = true;
            }
        }

        return {
            isCurrentlyExpanded: isCurrentlyExpanded,
            slideUpDuration: slideUpDuration,
            slideDownDuration: slideDownDuration,
        }
    }

    /**
    * FUNCTION: toBoolean
    * USE: Converts a specified value to an equivalent Boolean value.
    * @param value: Value to convert
    * @return: true if value evaluates to true; false otherwise.
    */
    let toBoolean = (value) => {
        if (typeof value === 'string') {
            value = value.trim().toLowerCase();
        }

        let ret = false;
        switch (value) {
            case true: case "true": case "yes": case "1": case 1:
                ret = true;
                break;
        }

        return ret;
    }
    
    (function (factory) {
        if (typeof define === 'function' && define.amd) {
            define([], factory);
        } else if (typeof exports === 'object') {
            module.exports = factory();
        }
    }(function() { return namespace; }));    
}(Collapsible)); // http://programmingnotes.org/